@@ -136,11 +136,17 @@ def sign_zipfile(file,key,hash="sha1",outfile=signedimp.HASHFILE_NAME):
     """Sign all the modules found in the given zipfile.
 
     This function walks the given zipfile looking for python modules, and
-    signs everything it finds in that file using the given key.
+    signs everything it finds in that file using the given key.  If the named
+    file is not a zipfile, the function returns immediately and leaves the
+    file unchanged.
 
     By default the signed hash file is written into the root of the zipfile;
     redirect output by passing a filename or file object as 'outfile'.
     """
+    try:
+        zipfile.ZipFile(file).close()
+    except zipfile.BadZipfile:
+        return
     infile = zipfile.ZipFile(file,"a")
     other_sigs = []
     #  Simulate "os" module for passing to the hash_files function.
@@ -535,7 +541,7 @@ else:
             #  Fortunately cxfreeze usually includes them as frozen modules
             #  directly into the exe; this is just to make sure.
             for nm2 in os.listdir(appdir):
-                if nm2 == nm:
+                if nm2 == nm or not os.path.isfile(os.path.join(appdir,nm2)):
                     continue
                 try:
                     zf2 = zipfile.ZipFile(os.path.join(appdir,nm2))
@@ -546,16 +552,18 @@ else:
                         for znm in zf2.namelist():
                             for incmod in ("encodings","codecs"):
                                 if znm.startswith(incmod):
-                                    zf.writestr(incmod,zf2.read(incmod))
+                                    zf.writestr(znm,zf2.read(znm))
                     finally:
                         zf2.close()
         finally:
             zf.close()
     #  Sign any zipfiles in the appdir (inlcuding the exes from above)
     for nm in os.listdir(appdir):
+        if not os.path.isfile(os.path.join(appdir,nm)):
+            continue
         try:
             sign_zipfile(os.path.join(appdir,nm),key,hash=hash)
-        except zipfile.BadZipfile:
+        except (zipfile.BadZipfile,EnvironmentError):
             pass
     #  Sign the main app dir.
     sign_directory(appdir,key,hash=hash)
